CREATE TYPE SSObject as ENUM (
    'Milky Way',
    'Sun',
    'Mercury',
    'Venus',
    'Earth',
    'Mars',
    'Jupiter',
    'Saturn',
    'Uranus',
    'Neptune',
    'Ceres',
    'Pluto',
    'Moon',
    'Phobos',
    'Deimos',
    'Io',
    'Europa',
    'Ganymede',
    'Callisto',
    'Mimas',
    'Enceladus',
    'Tethys',
    'Dione',
    'Rhea',
    'Titan',
    'Hyperion',
    'Iapetus',
    'Phoebe',
    'Charon',
    'Nix',
    'Hydra'
);
