import pytest
from nexoclom.utilities.configure import (configure, configure_nexoclom,
                                          configure_solarsystem,
                                          configure_atomicdata)

# @pytest.mark.utilities
# def test_configure():
